{* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *

Author:       francois.piette@overbyte.be, http://www.overbyte.be
Creation:     Jan 02, 2013
Description:  PersistAs attribute and handler class
              See accompanying demo for usage.
              Read blog article at http://francois-piette.blogspot.be/
Version:      1.00
History:


 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *}
unit AttrPersist;

interface

uses
    SysUtils, Classes, Rtti, TypInfo;

type
    PersistAs = class(TCustomAttribute)
    public
        Section  : String;
        Key      : String;
        DftValue : String;
        constructor Create(const ASection  : String;
                           const AKey      : String;
                           const ADftValue : String = '');
    end;

    TPersistState = (pstBeginUpdate,
                     pstReadField,
                     pstReadProp,
                     pstWriteField,
                     pstWriteProp,
                     pstEndUpdate);

    TPersistWriterProc = procedure (const Obj       : TObject;
                                    const State     : TPersistState;
                                    const ASection  : String;
                                    const AKey      : String;
                                    const AValue    : String) of object;
    TPersistReaderProc = procedure (const Obj       : TObject;
                                    const State     : TPersistState;
                                    const ASection  : String;
                                    const AKey      : String;
                                    const ADftValue : String;
                                    out   AValue    : String) of object;

    TPersistHandler = class
    protected
        class function  GetValue(var AValue : TValue) : String;
        class function  GetPersistAs(Obj : TRttiObject) : PersistAs;
        class procedure SetValue(var   AValue : TValue;
                                 const AData  : String);
    public
        class procedure Process(const Obj       : TObject;
                                const Writer : TPersistWriterProc);
                                overload;
        class procedure Process(const Obj       : TObject;
                                const Reader  : TPersistReaderProc);
                                overload;
    end;

    EPersistAs = class(Exception);

implementation

{ PersistAs }

constructor PersistAs.Create(const ASection, AKey, ADftValue: String);
begin
    Section  := ASection;
    Key      := AKey;
    DftValue := ADftValue;
end;

{ TPersistHandler }

class function TPersistHandler.GetPersistAs(Obj: TRttiObject): PersistAs;
var
    Attr : TCustomAttribute;
begin
    Result := nil;
    for Attr in Obj.GetAttributes do begin
        if Attr is PersistAs then begin
            Result := PersistAs(Attr);
            break;
        end;
    end;
end;

class function TPersistHandler.GetValue(var AValue: TValue): String;
begin
    if AValue.Kind in [tkWChar, tkLString, tkWString, tkString, tkChar,
        tkUString, tkInteger, tkInt64, tkFloat, tkEnumeration, tkSet] then
        Result := AValue.ToString
    else
        raise EPersistAs.Create(GetEnumName(TypeInfo(TTypeKind),
                                            Ord(AValue.Kind)) +
                                ': Type not Supported');
end;

class procedure TPersistHandler.SetValue(
    var   AValue : TValue;
    const AData  : String);
var
    NValue : Integer;
begin
    case AValue.Kind of
    tkWChar, tkLString, tkWString, tkString, tkChar, tkUString :
        AValue := AData;
    tkInteger, tkInt64 :
        AValue := StrToInt(AData);
    tkFloat :
        AValue := StrToFloat(AData);
    tkEnumeration :
        AValue := TValue.FromOrdinal(AValue.TypeInfo,
                                     GetEnumValue(AValue.TypeInfo, AData));
    tkSet :
        begin
            NValue := StringToSet(AValue.TypeInfo, AData);
            TValue.Make(@NValue, AValue.TypeInfo, AValue);
        end;
    else
        raise EPersistAs.Create(GetEnumName(TypeInfo(TTypeKind),
                                            Ord(AValue.Kind)) +
                                ': Type not Supported');
    end;
end;

class procedure TPersistHandler.Process(
    const Obj    : TObject;
    const Reader : TPersistReaderProc);
var
    RttiCtx    : TRttiContext;
    RttiType   : TRttiType;
    RttiField  : TRttiField;
    RttiProp   : TRttiProperty;
    AttrValue  : TValue;
    Attr       : PersistAs;
    AttrString : String;
begin
    if (not Assigned(Obj)) or (not Assigned(Reader)) then
        Exit;

    Reader(Obj, pstBeginUpdate, '', '', '', AttrString);
    try
        RttiCtx := TRttiContext.Create;
        try
            RttiType := RttiCtx.GetType(Obj.ClassInfo);
            for RttiProp in RttiType.GetProperties do begin
                Attr := GetPersistAs(RttiProp);
                if Assigned(Attr) then begin
                    AttrValue  := RttiProp.GetValue(Obj);
                    Reader(Obj, pstReadProp,
                           Attr.Section, Attr.Key, Attr.DftValue, AttrString);
                    SetValue(AttrValue, AttrString);
                    RttiProp.SetValue(Obj, AttrValue);
                end;
            end;
            for RttiField in RttiType.GetFields do begin
                Attr := GetPersistAs(RttiField);
                if Assigned(Attr) then begin
                    AttrValue  := RttiField.GetValue(Obj);
                    Reader(Obj, pstReadField,
                           Attr.Section, Attr.Key, Attr.DftValue, AttrString);
                    SetValue(AttrValue, AttrString);
                    RttiField.SetValue(Obj, AttrValue);
                end;
            end;
        finally
            RttiCtx.Free;
        end;
    finally
        Reader(Obj, pstEndUpdate, '', '', '', AttrString);
    end;
end;

class procedure TPersistHandler.Process(
    const Obj    : TObject;
    const Writer : TPersistWriterProc);
var
    RttiCtx     : TRttiContext;
    RttiType    : TRttiType;
    RttiField   : TRttiField;
    RttiProp    : TRttiProperty;
    AttrValue   : TValue;
    Attr        : PersistAs;
    AttrString  : String;
begin
    if (not Assigned(Obj)) or (not Assigned(Writer)) then
        Exit;

    Writer(Obj, pstBeginUpdate, '', '', '');
    try
        RttiCtx := TRttiContext.Create;
        try
            RttiType := RttiCtx.GetType(Obj.ClassInfo);
            for RttiProp in RttiType.GetProperties do begin
                Attr := GetPersistAs(RttiProp);
                if Assigned(Attr) then begin
                    AttrValue  := RttiProp.GetValue(Obj);
                    AttrString := GetValue(AttrValue);
                    Writer(Obj, pstWriteProp,
                           Attr.Section, Attr.Key, AttrString);
                end;
            end;
            for RttiField in RttiType.GetFields do begin
                Attr := GetPersistAs(RttiField);
                if Assigned(Attr) then begin
                    AttrValue  := RttiField.GetValue(Obj);
                    AttrString := GetValue(AttrValue);
                    Writer(Obj, pstWriteField,
                           Attr.Section, Attr.Key, AttrString);
                end;
            end;
        finally
            RttiCtx.Free;
        end;
    finally
        Writer(Obj, pstEndUpdate, '', '', '');
    end;
end;

end.
